@@ -38,7 +38,8 @@ public:
 	enum mode {
 		reading,
 		writing,
-		readwrite
+		readwrite,
+		appending,
 	};
 
 	/** \brief Creation flags when opening file for writing.
@@ -72,7 +73,10 @@ public:
 		 *
 		 * Does not modify permissions if the file already exists.
 		 */
-		 current_user_and_admins_only = 0x8
+		current_user_and_admins_only = 0x8,
+
+		/// Opening fails if the file already exists
+		fresh =  0x10
 	};
 
 	file() = default;
@@ -96,6 +100,11 @@ public:
 	bool opened() const;
 	explicit operator bool() const { return opened(); }
 
+	/*
+	 * \brief Opens the named file with the given mode
+	 *
+	 * If a (different) file was already opened, it is closed prior to opening the new file.
+	 */
 	result open(native_string const& f, mode m, creation_flags d = existing);
 
 	void close();
