@@ -5,10 +5,45 @@
 
 #include <algorithm>
 #include <cstdint>
+#include <limits>
+#include <optional>
 #include <string>
 #include <string_view>
 #include <vector>
 
+template<class CharT, class BaseT>
+class traits_cloner
+{
+public:
+	using char_type = CharT;
+
+	using base_type = BaseT;
+	using base_traits = std::char_traits<base_type>;
+
+	static std::size_t length(char_type const* s) {
+		return base_traits::length(reinterpret_cast<base_type const*>(s));
+	}
+	static int compare(char_type const* s1, char_type const* s2, std::size_t count) {
+		return base_traits::compare(reinterpret_cast<base_type const*>(s1), reinterpret_cast<base_type const*>(s2), count);
+	}
+	static char_type* copy(char_type* dest, char_type const* src, std::size_t count) {
+		return reinterpret_cast<char_type*>(base_traits::copy(reinterpret_cast<base_type*>(dest), reinterpret_cast<base_type const*>(src), count));
+	}
+	static void assign( char_type& c1, char_type const& c2 ) noexcept {
+	c1 = c2;
+	}
+	static char_type const* find(char_type const* ptr, std::size_t count, char_type const& ch) {
+		return reinterpret_cast<char_type const*>(base_traits::find(reinterpret_cast<base_type const*>(ptr), count, reinterpret_cast<base_type const&>(ch)));
+	}
+	static bool eq(char_type a, char_type b) {
+		return base_traits::eq(static_cast<base_type>(a), static_cast<base_type>(b));
+	}
+};
+
+template<>
+class std::char_traits<uint8_t> : public traits_cloner<uint8_t, char>
+{};
+
 /** \file
  * \brief String types and assorted functions.
  *
@@ -115,6 +150,35 @@ std::wstring FZ_PUBLIC_SYMBOL str_tolower_ascii(std::wstring_view const& s);
 std::string FZ_PUBLIC_SYMBOL str_toupper_ascii(std::string_view const& s);
 std::wstring FZ_PUBLIC_SYMBOL str_toupper_ascii(std::wstring_view const& s);
 
+/** \brief A type trait to identify the value_type of any string-like type.
+ */
+template<typename T, typename = void>
+struct string_value_type;
+
+template<typename T>
+struct string_value_type<T, std::void_t<typename T::value_type>> {
+	using type = typename T::value_type;
+};
+
+template<typename CharT>
+struct string_value_type<CharT*> {
+	using type = std::remove_cv_t<CharT>;
+};
+
+template<typename CharT, std::size_t N>
+struct string_value_type<CharT (&)[N]> {
+	using type = std::remove_cv_t<CharT>;
+};
+
+template<typename CharT, std::size_t N>
+struct string_value_type<CharT[N]> {
+	using type = std::remove_cv_t<CharT>;
+};
+
+// Helper alias for easier use.
+template<typename T>
+using string_value_type_t = typename string_value_type<T>::type;
+
 /** \brief Comparator to be used for std::map for case-insensitive keys
  *
  * Comparison is done locale-agnostic.
@@ -122,10 +186,16 @@ std::wstring FZ_PUBLIC_SYMBOL str_toupper_ascii(std::wstring_view const& s);
  */
 struct FZ_PUBLIC_SYMBOL less_insensitive_ascii final
 {
-	template<typename T>
-	bool operator()(T const& lhs, T const& rhs) const {
-		return std::lexicographical_compare(lhs.cbegin(), lhs.cend(), rhs.cbegin(), rhs.cend(),
-			[](typename T::value_type const& a, typename T::value_type const& b) {
+	using is_transparent = std::true_type;
+
+	template<typename L, typename R,
+		std::enable_if_t<std::is_same_v<string_value_type_t<L>, string_value_type_t<R>>>* = nullptr>
+	bool operator()(L const& lhs, R const& rhs) const {
+		std::basic_string_view<string_value_type_t<L>> lv(lhs);
+		std::basic_string_view<string_value_type_t<R>> rv(rhs);
+
+		return std::lexicographical_compare(lv.begin(), lv.end(), rv.begin(), rv.end(),
+			[](auto const& a, auto const& b) {
 				return tolower_ascii(a) < tolower_ascii(b);
 			}
 		);
@@ -136,17 +206,14 @@ struct FZ_PUBLIC_SYMBOL less_insensitive_ascii final
  *
  * Equivalent to str_tolower_ascii(a).compare(str_tolower_ascii(b));
  */
-inline bool equal_insensitive_ascii(std::string_view a, std::string_view b)
+template<typename A, typename B,
+	std::enable_if_t<std::is_same_v<string_value_type_t<A>, string_value_type_t<B>>>* = nullptr>
+inline bool equal_insensitive_ascii(A const& a, B const& b)
 {
-	return std::equal(a.cbegin(), a.cend(), b.cbegin(), b.cend(),
-		[](auto const& a, auto const& b) {
-			return tolower_ascii(a) == tolower_ascii(b);
-		}
-	);
-}
-inline bool equal_insensitive_ascii(std::wstring_view a, std::wstring_view b)
-{
-	return std::equal(a.cbegin(), a.cend(), b.cbegin(), b.cend(),
+	std::basic_string_view<string_value_type_t<A>> av(a);
+	std::basic_string_view<string_value_type_t<B>> bv(b);
+
+	return std::equal(av.cbegin(), av.cend(), bv.cbegin(), bv.cend(),
 		[](auto const& a, auto const& b) {
 			return tolower_ascii(a) == tolower_ascii(b);
 		}
@@ -507,55 +574,139 @@ inline auto FZ_PUBLIC_SYMBOL strtok_view(std::wstring_view const& tokens, wchar_
 
 /// \private
 template<typename T, typename String>
-T to_integral_impl(String const& s, T const errorval = T())
+bool to_integral_impl(String const& s, T & v)
 {
 	if constexpr (std::is_same_v<T, bool>) {
-		return static_cast<T>(to_integral_impl<unsigned int>(s, static_cast<unsigned int>(errorval))) != 0;
+		unsigned int w{};
+		if (!to_integral_impl(s, w)) {
+			return false;
+		}
+		v = w != 0;
+		return true;
 	}
 	else if constexpr (std::is_enum_v<T>) {
-		return static_cast<T>(to_integral_impl<std::underlying_type_t<T>>(s, static_cast<std::underlying_type_t<T>>(errorval)));
+		return to_integral_impl<std::underlying_type_t<T>>(s, reinterpret_cast<std::underlying_type_t<T>&>(v));
 	}
 	else {
-		T ret{};
+		bool negative{};
+
 		auto it = s.cbegin();
 		if (it != s.cend() && (*it == '-' || *it == '+')) {
+			if (*it == '-') {
+				if constexpr (std::is_signed_v<T>) {
+					negative = true;
+				}
+				else {
+					return false;
+				}
+			}
 			++it;
 		}
 
 		if (it == s.cend()) {
-			return errorval;
+			return false;
 		}
 
-		for (; it != s.cend(); ++it) {
-			auto const& c = *it;
-			if (c < '0' || c > '9') {
-				return errorval;
+		v = T{};
+		if (negative) {
+			auto constexpr min = std::numeric_limits<T>::min();
+			auto constexpr min10 = min / 10;
+			for (; it != s.cend(); ++it) {
+				auto const& c = *it;
+				if (c < '0' || c > '9') {
+					return false;
+				}
+				if (v < min10) {
+					return false;
+				}
+				v *= 10;
+				auto digit = -static_cast<T>(c - '0');
+				if (min - v > digit) {
+					return false;
+				}
+				v += digit;
 			}
-			ret *= 10;
-			ret += c - '0';
 		}
-
-		if (!s.empty() && s.front() == '-') {
-			ret *= static_cast<T>(-1);
+		else {
+			auto constexpr max = std::numeric_limits<T>::max();
+			auto constexpr max10 = max / 10;
+			for (; it != s.cend(); ++it) {
+				auto const& c = *it;
+				if (c < '0' || c > '9') {
+					return false;
+				}
+				if (v > max10) {
+					return false;
+				}
+				v *= 10;
+				auto digit = static_cast<T>(c - '0');
+				if (max - v < digit) {
+					return false;
+				}
+				v += digit;
+			}
 		}
-		return ret;
 	}
+	return true;
 }
 
 /// Converts string to integral type T. If string is not convertible, errorval is returned.
 template<typename T>
 T to_integral(std::string_view const& s, T const errorval = T()) {
-	return to_integral_impl<T>(s, errorval);
+	T out{};
+	if (!to_integral_impl<T>(s, out)) {
+		out = errorval;
+	}
+	return out;
 }
 
 template<typename T>
 T to_integral(std::wstring_view const& s, T const errorval = T()) {
-	return to_integral_impl<T>(s, errorval);
+	T out{};
+	if (!to_integral_impl<T>(s, out)) {
+		out = errorval;
+	}
+	return out;
 }
 
 template<typename T, typename StringType>
 T to_integral(std::basic_string_view<StringType> const& s, T const errorval = T()) {
-	return to_integral_impl<T>(s, errorval);
+	T out{};
+	if (!to_integral_impl<T>(s, out)) {
+		out = errorval;
+	}
+	return out;
+}
+
+/// Converts string to integral type T. If string is not convertible, nullopt
+template<typename T>
+std::optional<T> to_integral_o(std::string_view const& s) {
+	std::optional<T> ret;
+	T out{};
+	if (to_integral_impl<T>(s, out)) {
+		ret = out;
+	}
+	return ret;
+}
+
+template<typename T>
+std::optional<T> to_integral_o(std::wstring_view const& s) {
+	std::optional<T> ret;
+	T out{};
+	if (to_integral_impl<T>(s, out)) {
+		ret = out;
+	}
+	return ret;
+}
+
+template<typename T, typename StringType>
+std::optional<T> to_integral_o(std::basic_string_view<StringType> const& s) {
+	std::optional<T> ret;
+	T out{};
+	if (to_integral_impl<T>(s, out)) {
+		ret = out;
+	}
+	return ret;
 }
 
 
@@ -669,19 +820,23 @@ inline void rtrim(String & s, std::wstring_view const& chars = L" \r\n\t")
  *
  * \tparam insensitive_ascii If true, comparison is case-insensitive
  */
-template<bool insensitive_ascii = false, typename String>
-bool starts_with(String const& s, String const& beginning)
+template<bool insensitive_ascii = false, typename String, typename Beginning>
+	std::enable_if_t<std::is_same_v<string_value_type_t<String>, string_value_type_t<Beginning>>,
+bool> starts_with(String const& s, Beginning const& beginning)
 {
-	if (beginning.size() > s.size()) {
+	std::basic_string_view<string_value_type_t<String>> sv(s);
+	std::basic_string_view<string_value_type_t<Beginning>> beginningv(beginning);
+
+	if (beginningv.size() > sv.size()) {
 		return false;
 	}
+
 	if constexpr (insensitive_ascii) {
-		return std::equal(beginning.begin(), beginning.end(), s.begin(), [](typename String::value_type const& a, typename String::value_type const& b) {
+		return std::equal(beginningv.begin(), beginningv.end(), sv.begin(), [](auto const& a, auto const& b) {
 			return tolower_ascii(a) == tolower_ascii(b);
 		});
-	}
-	else {
-		return std::equal(beginning.begin(), beginning.end(), s.begin());
+	} else {
+		return std::equal(beginningv.begin(), beginningv.end(), sv.begin());
 	}
 }
 
@@ -689,20 +844,24 @@ bool starts_with(String const& s, String const& beginning)
  *
  * \tparam insensitive_ascii If true, comparison is case-insensitive
  */
-template<bool insensitive_ascii = false, typename String>
-bool ends_with(String const& s, String const& ending)
+template<bool insensitive_ascii = false, typename String, typename Ending>
+	std::enable_if_t<std::is_same_v<string_value_type_t<String>, string_value_type_t<Ending>>,
+bool> ends_with(String const& s, Ending const& ending)
 {
-	if (ending.size() > s.size()) {
+	std::basic_string_view<string_value_type_t<String>> sv(s);
+	std::basic_string_view<string_value_type_t<Ending>> endingv(ending);
+
+	if (endingv.size() > sv.size()) {
 		return false;
 	}
 
 	if constexpr (insensitive_ascii) {
-		return std::equal(ending.rbegin(), ending.rend(), s.rbegin(), [](typename String::value_type const& a, typename String::value_type const& b) {
+		return std::equal(endingv.rbegin(), endingv.rend(), sv.rbegin(), [](auto const& a, auto const& b) {
 			return tolower_ascii(a) == tolower_ascii(b);
 		});
 	}
 	else {
-		return std::equal(ending.rbegin(), ending.rend(), s.rbegin());
+		return std::equal(endingv.rbegin(), endingv.rend(), sv.rbegin());
 	}
 }
 
@@ -779,6 +938,8 @@ inline native_string to_native_from_utf8(std::string_view s) {
 #endif
 }
 
+FZ_PUBLIC_SYMBOL void wipe_conversion_cache();
+
 }
 
 #endif
